#Robot Personality Simulator
'''created by Casey Bennett 2019, www.CaseyBennett.com
   Copyright 2019

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License (Lesser GPL) as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.
'''
# For multiprocessing, this version varies as follow:
#   1) The command line user input commented out
#   2) Main takes input parameters, so it can be called directly
#   3) All output file writing is commented out

import os
import sys
import math
import random
import time
from optparse import OptionParser
import numpy as np

import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from matplotlib import cm

from keras.models import Sequential, Model
from keras.layers import Input, Dense, Activation, Dropout, concatenate
import tensorflow as tf

import cfg
import nn

#Suppress tensorflow warnings
#tf.logging.set_verbosity(tf.logging.ERROR)                     #Doesn't work in tensflow 2.0 and up, use below lines
import logging
logger = tf.get_logger()
logger.setLevel(logging.ERROR)


#############################################################################
#
#User Input Program Parameters
#
#######################

######### Variable Parameters ##############
"""This is used to switch various option levels below, altering
    the functioning methodology of the program in order to
    test differenet scenarios (SEE ReadMe)"""
cfg.local=cfg.congruent=cfg.nn_type=cfg.agent_goal_type=sys.maxsize
cfg.env_learn_type=cfg.env_noise=cfg.agent_move_type=sys.maxsize


'''#Command Line Switch - Check First
parser = OptionParser()
(options, args) = parser.parse_args()

#If #args=#switches, then load, else skip to keyboard prompt below
if len(args)==7:    
    cfg.local=int(args[0])                  #Sets local parameter
    cfg.congruent=int(args[1])              #Sets congruent parameter
    cfg.nn_type=int(args[2])                #Sets nn_type parameter
    cfg.agent_goal_type=int(args[3])        #Sets agent_goal_type parameter
    cfg.env_learn_type=int(args[4])         #Sets env_learn_type parameter, use Sensed States or Sensed Deltas
    cfg.env_noise=float(args[5])              #Sets the level of environmental noise
    cfg.agent_move_type=int(args[6])        #Sets agent_move_type parameter

#Keyboard Switch Prompt
while cfg.local>2:
    cfg.local = int(input("Local Switch (0,1,2 ... SEE ReadMe): "))                         
while cfg.congruent>2:
    cfg.congruent = int(input("Congruent Switch (0,1,2 ... SEE ReadMe): "))
while cfg.nn_type>2:
    cfg.nn_type = int(input("NN Type Switch (1,2 ... SEE ReadMe): "))
while cfg.agent_goal_type>2:
    cfg.agent_goal_type = int(input("Agent Goal Type Switch (0,1,2 ... SEE ReadMe): "))
while cfg.env_learn_type>2:
    cfg.env_learn_type = int(input("Env Learn Type, controls if agent used states or deltas for learning (1,2 ... SEE ReadMe): "))
while cfg.env_noise>1:
    cfg.env_noise = float(input("Environmental Noise setting (value between 0 and 1 ... SEE ReadMe): "))
while cfg.agent_move_type>2:
    cfg.agent_move_type = int(input("Agent Move Type Switch (1,2 ... SEE ReadMe): "))
'''


#Variables for storing final output
init_error=None
final_error=None
converge_step=cfg.max_iter
deconverged=0
deconverge_cnt=0


##############################################################################
#
# Define Classes
#
######################

 # 1) need agent class
 # 2) Agent class should have functions for emotion-behavior model
 # 3) also need autoencoder

class Agent:
    def __init__(self):
        self.row_cnt=0

    def get_row_cnt():
        return row_cnt

class Robot(Agent):    
    #Robot_ID
    #curr_loc=[]
    #curr_emot
    #curr_act
    #next_emot
    #focal_pt
    #dull_ctr
    #action_hist = []
    #sensed_states = []
    #sensed_deltas = []
    #NN_model
    #NN_results
    
    def __init__(self, n, loc_init):
        self.Robot_ID = n
        self.curr_loc=loc_init
        """Should be 2 values specifying [0]=x loc and [1]=y loc"""
        self.curr_emot = random.uniform(0.45,0.55)
        """Emotional state is defined on a 0-1 scale, initially randomized around neutral (0.5)"""
        self.curr_act=None
        """Current action"""
        self.next_emot=[]
        """Predicted next emotion, for each action"""
        self.focal_pt=None
        """Focal point of attention"""
        self.dull_ctr=0
        """Counter for dulling effect, can be used to speed up or slow down effect"""
        self.action_hist = []
        """Array with each action agent performed over time"""
        self.sensed_states = [[0.5 for col in range(cfg.env_height)] for row in range(cfg.env_width)]
        '''Sensed state of each point in environment grid'''
        self.sensed_states_hist = [[0.5 for col in range(cfg.env_height)] for row in range(cfg.env_width)]
        '''Previous Sensed state of each point in environment grid'''
        self.sensed_deltas = [[0 for col in range(cfg.env_height)] for row in range(cfg.env_width)]
        '''Sensed delta of each point in environment grid'''
        self.sensed_deltas_hist = [[0 for col in range(cfg.env_height)] for row in range(cfg.env_width)]
        '''Previous Sensed delta of each point in environment grid'''
        self.stimuli = []
        '''Array for attention stimuli [X,Y,area,decay,time_out,st_found,attend_bin] ... CURRENTLY UNUSED'''
        self.NN_model=None
        self.NN_create()
        """Neural Network model"""
        self.NN_results=[]
        """Holds training results for NN, for later access if needed"""   


    def Choose_Action(self):
        '''Picks action based on emotion state and current environment, via the NN.
            Note for the "no goal" scenario, just uses the current emotion state, since
            w/o goal no way to pick among multiple predicted next emotions'''

        #First setup vars for next steps
        emot_diff=0
        action = random.randint(0,len(cfg.action_list)-1)
        split = 1.0 / len(cfg.action_list)

        #Determine optimal action, based on agent goal type
        if cfg.agent_goal_type==0:                                                                      #The "no goal" version, just picks action based on curr emot, using some basic logic

            if self.curr_emot==0:
                action=0                                                                                #special case, when curr emot is 0

            elif self.curr_emot>0:                                                                      #else run logic
                for i in range(len(cfg.action_list)):
                    if self.curr_emot > split * i:                                                          #lower threshold
                        if self.curr_emot <= split * (i+1):                                                 #upper threshold
                            action=i
 
        elif cfg.agent_goal_type==1:                                                                    #Max positive emotion goal 
            for i in range(len(cfg.action_list)):
                test = self.next_emot[i] - self.curr_emot
                if test>emot_diff:
                    emot_diff=test
                    action=i

        elif cfg.agent_goal_type==2:                                                                    #Max total emotion (diff from neutral) goal
            for i in range(len(cfg.action_list)):
                test = abs(self.next_emot[i] - 0.5)
                if test>emot_diff:
                    emot_diff=test
                    action=i


        #print('YOLOOOO:', emot_diff, action)

                           
        #Set new action
        self.curr_act=action
        self.action_hist.append(action)


    def Predict_Emot(self):
        '''Predicts next emotion using autoencoder'''

        #Make prediction, depending on nn_type
        if cfg.nn_type==1:

            #Get correct env info first
            if cfg.env_learn_type==1:
                env_data=self.sensed_states
            elif cfg.env_learn_type==2:
                env_data=self.sensed_deltas
            
            pred = nn.MLP_predict(self.NN_model, cfg.action_list, env_data)
            self.next_emot = pred.reshape(len(cfg.action_list)).tolist()                                #Reshape prediction numpy array and convert into native python list


        if cfg.nn_type==2:
            
            #Get correct env info first
            env_data = [[[0.5 for col in range(cfg.env_height)] for row in range(cfg.env_width)] for t in range(2)]
            if cfg.env_learn_type==1:
                env_data[0]=self.sensed_states_hist
                env_data[1]=self.sensed_states
            elif cfg.env_learn_type==2:
                env_data[0]=self.sensed_deltas_hist
                env_data[1]=self.sensed_deltas
                
            pred = nn.RNN_predict(self.NN_model, cfg.action_list, env_data)
            self.next_emot = pred.reshape(len(cfg.action_list)).tolist()                                #Reshape prediction numpy array and convert into native python list


    def Update_Emot(self):                                                                          #!!!!Perhaps consider having attention span variables learned by agent in future??? 
        '''Updates agent emotional state'''
                                                                                                 
        ####Calc Env Delta and attention first########
        curr_delta=self.Calc_Env_Deltas()
        self.Attention_Decay(curr_delta)
        
        ####Taking into account attention, update emotion########
        if self.dull_ctr>=cfg.att_span:
            self.curr_emot += (0.5-self.curr_emot)/cfg.att_decay_rate                               #If past attention span, emotion to decays back to neutral
        else:
            self.curr_emot=max(min(self.curr_emot+curr_delta,1),0)                                  #Else, update emotion based on env delta


        ####Completely random emotion OVERRIDE, just for baseline testing########
        if cfg.emot_rand==1:
            self.curr_emot=random.random()


    def Sense_Env(self, env):
        '''Senses current environment and holds info, also detects
            environmental changes (i.e. deltas) during process'''
        
        for i in range(len(env.states)):
            for j in range(len(env.states[0])):
                self.sensed_deltas[i][j] = env.states[i][j] - self.sensed_states[i][j]              #Calcs deltas first 
                self.sensed_states[i][j] = env.states[i][j]                                         #Then updates sensed env array


    def Calc_Env_Deltas(self):
        '''Calc Env Deltas, returns curr delta. Varies based on local parameter'''
        
        #Global
        if cfg.local==0:            
            curr_delta = np.mean(self.sensed_deltas)
            area=cfg.env_height*cfg.env_width

        #Local, loop thru local environment   
        elif cfg.local==1:            
            curr_delta=ctr=0
            for i in range(-1,2):
                for j in range(-1,2):
                    curr_x=self.curr_loc[0]+i
                    curr_y=self.curr_loc[1]+j
                    if curr_x>=0 and curr_x<cfg.env_width and curr_y>=0 and curr_y<cfg.env_height:
                        ctr+=1
                        curr_delta += self.sensed_deltas[curr_x][curr_y]

            curr_delta = round(float(curr_delta)/ctr,5)
            area=9
            
        #Super Local, current location only
        elif cfg.local==2:            
            curr_delta = self.sensed_deltas[self.curr_loc[0]][self.curr_loc[1]]
            area=1

        return curr_delta           

            
    def Move(self):
        '''Moves agent towards focal point of attention, depending
            on how agent_move_type is set (e.g. direct look, biased
            saccades, etc.)'''                                          

        #Calc distance to focal point
        dist = math.sqrt(float(self.curr_loc[0]-self.focal_pt[0])**2 + float(self.curr_loc[1]-self.focal_pt[1])**2)
        dist_x = self.curr_loc[0]-self.focal_pt[0]
        dist_y = self.curr_loc[1]-self.focal_pt[1]

        #Calc movement
        new_x=new_y=0                                                                               #Start assuming no motion, so =0
        
        if cfg.agent_move_type==1:                                                                  #Direct Look
            new_x = self.curr_loc[0]-dist_x
            new_y = self.curr_loc[1]-dist_y
            
        elif cfg.agent_move_type==2:                                                                #Biased saccades
            if (dist_x>0):
                new_x=self.curr_loc[0]-random.randint(0,dist_x)
            elif (dist_x<0):
                new_x=self.curr_loc[0]-random.randint(dist_x,0)
            if (dist_y>0):
                new_y=self.curr_loc[1]-random.randint(0,dist_y)
            elif (dist_y<0):
                new_y=self.curr_loc[1]-random.randint(dist_y,0)

        #Limits to new location to max/min of environment
        new_x=max(min(new_x,cfg.env_width-1),0)
        new_y=max(min(new_y,cfg.env_height-1),0)

        #Agent "moves" to new location
        self.curr_loc=[new_x,new_y]                                                          


    def Find_Focal(self):
        '''Finds max state value of env grid, this can be thought of as equivalent to
            localizing the focal point of attention (e.g. based on optical flow) during
            human-robot interaction ... may need to be replaced with a full blown
            Calc_Attention function for multiple stimuli at some point'''

        #Find Focal point
        max_val=-sys.maxsize
        max_loc=[]
        for i in range(len(self.sensed_states)):
            for j in range(len(self.sensed_states[0])):
                if self.sensed_states[i][j]>max_val:
                    max_val=self.sensed_states[i][j]
                    max_loc=[i,j]

        self.focal_pt=max_loc

        #Create stimuli
        area=1
        self.stimuli = [self.curr_loc[0], self.curr_loc[1], area, 1, cfg.att_span, 1, 1]            #Array for attention stimuli [np,nt,area,decay,time_out,st_found,attend_bin] ... CURRENTLY UNUSED, just a placeholder
    
        return(max_loc, max_val) 


    def Attention_Decay(self, curr_delta):
        '''Calc attention decay of agent based on current stimuli, using dulling effect to account
            for attention span and IOR (inhibition of return) in natural organisms ... this eventually may
            need to be altered to match attention functions and stimuli in RobotFace.py'''
        
        #Calc dull ctr
        if abs(curr_delta)<cfg.dull_thresh:                                                         #If abs val of curr_delta less than dulling threshold
            self.dull_ctr += 1
        else:
            self.dull_ctr=0       


    def NN_create(self):
        '''Constructs the NN for the agent, depending on nn_type, to learn action-emotion model
            of how the world responds to it'''

        if cfg.nn_type==1:
            self.NN_model=nn.MLP_construct()
        elif cfg.nn_type==2:
            self.NN_model=nn.RNN_construct()
            

    def NN_learn(self):
        '''Updates the NN, depending on nn_type, based on how the world responds to it (env deltas),
            given its last action, as measured by its internal emotion'''    


        #Update NN, depending on nn_type
        if cfg.nn_type==1:

            #Get corrrect env info first
            if cfg.env_learn_type==1:
                env_data=self.sensed_states
            elif cfg.env_learn_type==2:
                env_data=self.sensed_deltas
            
            result=nn.MLP_update(self.NN_model, self.curr_emot, self.curr_act, env_data)
            self.NN_results.append(result)

            
        elif cfg.nn_type==2:

            #Get corrrect env info first
            env_data = [[[0.5 for col in range(cfg.env_height)] for row in range(cfg.env_width)] for t in range(2)]
            if cfg.env_learn_type==1:
                env_data[0]=self.sensed_states_hist
                env_data[1]=self.sensed_states
            elif cfg.env_learn_type==2:
                env_data[0]=self.sensed_deltas_hist
                env_data[1]=self.sensed_deltas
            
            result=nn.RNN_update(self.NN_model, self.curr_emot, self.curr_act, env_data)
            self.NN_results.append(result)


        
class Env():
    #states=[]

    def __init__(self):
        self.states = [[0.5 for col in range(cfg.env_height)] for row in range(cfg.env_width)]
        '''State of each point in grid'''

    def Update_Env(self, action, loc):
        '''Updates environment based on agent action and location'''                                
        
        #Incongruent response     
        if cfg.congruent==0:
            if action==0:
                delta = 1 * cfg.env_step_size
            elif action==1:
                delta = random.randint(-1,1) * cfg.env_step_size 
            elif action==2:
                delta = -1 * cfg.env_step_size
        #Congruent response
        elif cfg.congruent==1:
            if action==2:
                delta = 1 * cfg.env_step_size
            elif action==1:
                delta = random.randint(-1,1) * cfg.env_step_size 
            elif action==0:
                delta = -1 * cfg.env_step_size

        #Random response
        elif cfg.congruent==2:
            delta = random.randint(-1,1) * cfg.env_step_size

        #Add some noise
        delta_noise = delta * cfg.env_noise                                                             #Added noise to environmental changes based on env_rand parameter

        #Loop thru env grid, update env states
        for i in range(len(self.states)):
            for j in range(len(self.states[0])):
                dist = math.sqrt(float(loc[0]-i)**2 + float(loc[1]-j)**2)                               #Distance from current location, as euclidean distance
                new_val = (delta *(1/(dist+1)))                                                         #Delta, scaled by inverse distance
                new_val += random.uniform(-delta_noise,delta_noise)                                     #Add some noise in
                #print(delta, ":", dist, ":", new_val + self.states[i][j])
                self.states[i][j]=max(min(self.states[i][j]+new_val, 1.0),0.0)                          #Update State, min-max range       
               


                  
#############################################################################
#
# Main Program
#
#######################

##### Open Specifics Outfile ################                                                               
#cfg.outfile_sp.write("Step" + "\t" + "Current Emotion" + "\t" + "Current Action" + "\t" + "Current Location" + "\t" + "Dull Ctr" + "\t" + "Focal Val" + "\t" + "Focal Val" + "\n")

def main(local,congruent,nn_type,agent_goal_type,env_learn_type,env_noise,agent_move_type):

    #Start timer for runtime
    start_time = time.time()
    
    # Start the program
    cfg.local=local
    cfg.congruent=congruent
    cfg.nn_type=nn_type
    cfg.agent_goal_type=agent_goal_type
    cfg.env_learn_type=env_learn_type
    cfg.env_noise=env_noise
    cfg.agent_move_type=agent_move_type
    
    #Setup Environment
    Env1=Env()
    
    #Create agent object
    loc_init=[int(cfg.env_width/2), int(cfg.env_height/2)]                                              #Initializes start to center location (approximate)
    R1=Robot(1,loc_init)
    #print('Start Emotion:', R1.curr_emot)

    #Initialize sensed environment in agent
    R1.Sense_Env(Env1)

    #Main Loop
    for i in range(cfg.max_iter):

        #Agent Move (skip first iter)
        if i>0:
            focal_pt, focal_val=R1.Find_Focal()                                                     #Identify focal point of attention, robot/agent will move in this direction
            R1.Move()
        else:
            focal_val=0.5                                                                           #Environment initializes at 0.5
        #Agent predict next emotion (based on current environment and various actions)
        R1.Predict_Emot()
        #Agent does Stuff
        R1.Choose_Action()
        #Environment responds
        Env1.Update_Env(R1.curr_act, R1.curr_loc)
        #Agent senses environmental change, updates emotions
        R1.Sense_Env(Env1)
        R1.Update_Emot()
        #Autoencoder learning                                                                       
        R1.NN_learn()
        
        
        #Write Outfile Specifics
        #print('output_sp:', i, round(R1.curr_emot,4), R1.curr_act, R1.curr_loc, R1.dull_ctr, round(focal_val,4),R1.next_emot)
        #cfg.outfile_sp.write(str(i) + "\t" + str(round(R1.curr_emot,4)) + "\t" + str(R1.curr_act) + "\t" + str(R1.curr_loc) + "\t" + str(R1.dull_ctr) + "\t" + \
        #                     str(round(focal_val,4)) + "\t" + str(R1.next_emot) + "\n")


    #Test Print
    #print('\n', R1.action_hist)

    ##### Plot Epoch training & validation loss values########    
    plt.plot(R1.NN_results)
    plt.title('Model loss')
    plt.ylabel('Loss')
    plt.xlabel('Epoch')
    plt.legend(['Train', 'Test'], loc='upper left')
    plt.savefig('NN_Training_History.png', dpi=300)
    #plt.show()
    

    #Write out NN learning results by epoch
    prev=sys.maxsize
    global converge_step, deconverged, deconverge_cnt
    converge_step=cfg.max_iter
    deconverged=0
    deconverge_cnt=0
    
    for i in range(len(R1.NN_results)):
        #cfg.outfile_nn_learning.write(str(i) + "\t" + str(round(R1.NN_results[i],6)) + "\t" + str(round(abs(prev-R1.NN_results[i]),6)) + "\n")

        #For tracking convergence stats
        if abs(prev-R1.NN_results[i])<0.001 and converge_step==cfg.max_iter:
            converge_step=i
        if abs(prev-R1.NN_results[i])>0.001 and converge_step<cfg.max_iter:
            converge_step=cfg.max_iter
            deconverged=1
            deconverge_cnt+=1
        prev=R1.NN_results[i]
            
    #Save values for final output
    global init_error, final_error
    init_error=R1.NN_results[0]
    final_error=R1.NN_results[len(R1.NN_results)-1]

    prog_run_time=time.time()-start_time
    return(round(init_error,6),round(final_error,6),converge_step,deconverged,deconverge_cnt,prog_run_time)
  

################################################################################
'''
# Execute the simulation.
if __name__ == '__main__':
    main()

##### Open Main Outfile and Write Results ################
outfile = open("outfile.txt", 'w')                                                                  #!!!!Update Outfile

#Single line, all calculations, for Controller to use
outfile.write("Initial Error" + "\t" + "Final Error" + "\t" + "Converge Step" + "\t" + "Deconverged" + "\t" + "Deconverge Cnt" + "\n")
outfile.write(str(round(init_error,6)) + "\t" + str(round(final_error,6)) + "\t" + str(converge_step) + "\t" + str(deconverged) + "\t" + str(deconverge_cnt) + "\n")

#Write all the above calcations to main outfile separately
outfile.write("\nProgram Run Time:" + "\t" + str(time.time()-start_time) + "\n")
outfile.write("Epoch Cnt:" + "\t" + str(cfg.max_iter) + "\n")
outfile.write("Initial Error" + "\t" + str(round(init_error,6)) +  "\n")
outfile.write("Final Error" + "\t" + str(round(final_error,6)) +  "\n")
outfile.write("Converge Step" + "\t" + str(converge_step) +  "\n")
outfile.write("Deconverged" + "\t" + str(deconverged) +  "\n")
outfile.write("Deconverge Cnt" + "\t" + str(deconverge_cnt) +  "\n")


##### Close outfiles ################
outfile.close()
cfg.outfile_sp.close()
cfg.outfile_nn_learning.close()
'''
