#Runs multiple simulations via prod_sim.py using run parameters in infile
#This version uses multiprocessing 'Pool' to limit CPU usage

import sys
import csv
import os
import time
import numpy
import matplotlib.pyplot as pyplot
from mpl_toolkits.mplot3d import Axes3D
from matplotlib import cm

import multiprocessing
import Multiproc_RoboPers_Sim

param_cnt=7
field_cnt=16


##################################
#
# Main Helper Function
#
####################
'''Use this for calling multiple procs in __main__ below'''
def worker(local,congruent,nn_type,agent_goal_type,env_learn_type,env_noise,agent_move_type, model_sort, repeat):

    print(multiprocessing.current_process().name, 'Starting')

    #Run python program    
    output_data = Multiproc_RoboPers_Sim.main(local,congruent,nn_type,agent_goal_type,env_learn_type,env_noise,agent_move_type)
    out_arr=[model_sort, repeat, output_data]

    print(multiprocessing.current_process().name, 'Exiting')
    return out_arr
    

######################################################################
#
# Main Loop - Run Program for each Run_Param setup
#
###########################

#Global output array
output=[]                          

#Create header
header=['Model Sort','Local Switch','Congruent Switch','NN Type','Agent Goal Type','Env Learn Type','Env Noise','Agent Move Type',\
                'Init Error','Final Error','Converge Step','Deconverged','Deconverge Cnt', 'Program RunTime', 'Run Params', 'Repeat']

def main():

    #Setup Multiproc
    p = multiprocessing.Pool(4)
    args_input=[]
    results=[]
    mp_start=time.time()

    #Load Infile
    run_params = []                                                                       
    infile= csv.reader(open('infile.csv'), delimiter=',', quotechar='"')
    infile.__next__()
    for row in infile:
        run_params.append(row)

    #Set repeat cnt for probabilistic models (for averaging over sample)
    tot_reps=30

    ##### Run prod_sim.py using params ########
    for i in range(len(run_params)):
        print('---- Run_Param Setup', i+1, '... out of', len(run_params), '--------')
        local=int(run_params[i][0])                 #Sets local parameter
        congruent=int(run_params[i][1])             #Sets congruent parameter
        nn_type=int(run_params[i][2])               #Sets nn_type parameter
        agent_goal_type=int(run_params[i][3])       #Sets agent_goal_type parameter
        env_learn_type=int(run_params[i][4])        #Sets env_learn_type parameter, use Sensed States or Sensed Deltas
        env_noise=float(run_params[i][5])           #Sets the level of environmental noise
        agent_move_type=int(run_params[i][6])       #Sets agent_move_type parameter

        #Create Run String (using multiproc worker)
        '''this is defined in Main Helper worker function above (for multiprocessing), rather than in each loop'''

        #Run program, multiproc
        repeat=tot_reps
        while repeat>0:
            print("*Repeats Left:", repeat)            

            ####Update Repeat Value############
            repeat-=1

            #Create args
            args_input.append((local,congruent,nn_type,agent_goal_type,env_learn_type,env_noise,agent_move_type, i, repeat))            #Do this in loop, then call multiproc outside loop for simultaneous runs

    #Run Multiproc
    result = [p.apply_async(worker, args=x) for x in args_input]
    results = [x.get() for x in result]

    #Ensure we wait for all worker processes to finish
    p.close()
    p.join()  
        
    print('\n\n', results[0:5])
    print('output cnt:', len(results))
    print('MP runtime:', str(time.time()-mp_start))
            

    ##### Output Array Load ##########
    for i in range(len(run_params)):
        #print('---- Run_Param Setup', i+1, '... out of', len(run_params), '--------')
        local=int(run_params[i][0])                 #Sets local parameter
        congruent=int(run_params[i][1])             #Sets congruent parameter
        nn_type=int(run_params[i][2])               #Sets nn_type parameter
        agent_goal_type=int(run_params[i][3])       #Sets agent_goal_type parameter
        env_learn_type=int(run_params[i][4])        #Sets env_learn_type parameter, use Sensed States or Sensed Deltas
        env_noise=float(run_params[i][5])           #Sets the level of environmental noise
        agent_move_type=int(run_params[i][6])       #Sets agent_move_type parameter


        for j in range(len(results)):
            if results[j][0]==i:
                repeat=results[j][1]
                output_data=results[j][2]
                          
                #Set Descriptors
                temp_array2=[]
                temp_array2.append(i)                                   #Just use the run order as "model sort" in this version

                if local==0:
                    temp_array2.append('Global')
                elif local==1:
                    temp_array2.append('Local')
                elif local==2:
                    temp_array2.append('Super-Local')
                    
                if congruent==0:
                    temp_array2.append('Incongruent')
                elif congruent==1:
                    temp_array2.append('Congruent')
                elif congruent==2:
                    temp_array2.append('Random')

                if nn_type==1:
                    temp_array2.append('MLP')
                elif nn_type==2:
                    temp_array2.append('RNN')

                if agent_goal_type==0:
                    temp_array2.append('None')
                elif agent_goal_type==1:
                    temp_array2.append('Max Pos')
                elif agent_goal_type==2:
                    temp_array2.append('Diff Neutral')

                if env_learn_type==1:
                    temp_array2.append('States')
                elif env_learn_type==2:
                    temp_array2.append('Deltas')

                temp_array2.append(env_noise)

                if agent_move_type==1:
                    temp_array2.append('Direct Look')
                elif agent_move_type==2:
                    temp_array2.append('Biased Saccades')


                #Insert Output Data
                for x in range(len(output_data)):
                    temp_array2.append(output_data[x])

                #Insert Prog RunTime
                #temp_array2.append(prog_run_time)                                      #Handled in output data step above, in this version

                #Insert Run Params
                run_param_str=('"'+ str(local) + ',' + str(congruent) + ',' + str(nn_type) + ',' + str(agent_goal_type) + ',' +  \
                               str(env_learn_type) + ',' + str(env_noise) + ',' + str(agent_move_type) + '"')
                temp_array2.append(run_param_str)

                #Insert Repeat Value
                temp_array2.append(repeat)

                #Finally, load it into output array
                output.append(temp_array2)

    return(header,output)


#Run main    
if __name__ == '__main__':
    #multiprocessing.freeze_support()        #here if program needs to be frozen
    main()                                  # execute this only when run directly, not when imported!
        

##################################################################
#
# Write Results 
#
##################

#Open Results outfile
results_file=open("results.csv", 'w')

#Write Header
for i in range(field_cnt):
    results_file.write(str(header[i]) + ',')
results_file.write('\n')

#Write Data
for i in range(len(output)):
    for j in range(field_cnt):
        results_file.write(str(output[i][j]) + ',')
    results_file.write('\n')

results_file.close()


