#Runs multiple simulations via prod_sim.py using run parameters in infile

import sys
import csv
import os
import numpy
import matplotlib.pyplot as pyplot
from mpl_toolkits.mplot3d import Axes3D
from matplotlib import cm

param_cnt=7
field_cnt=16


######################################################################
#
# Main Loop - Run Program for each Run_Param setup
#
###########################

#Output array
output=[]

#Load Infile
run_params = []                                                                       
infile= csv.reader(open('infile.csv'), delimiter=',', quotechar='"')
infile.__next__()
for row in infile:
    run_params.append(row)

#Create header
header=['Model Sort','Local Switch','Congruent Switch','NN Type','Agent Goal Type','Env Learn Type','Env Noise','Agent Move Type',\
                           'Init Error','Final Error','Converge Step','Deconverged','Deconverge Cnt', 'Program RunTime', 'Run Params', 'Repeat']

#Run prod_sim.py using params
for i in range(len(run_params)):
    print('---- Run_Param Setup', i+1, '... out of', len(run_params), '--------')
    local=int(run_params[i][0])                 #Sets local parameter
    congruent=int(run_params[i][1])             #Sets congruent parameter
    nn_type=int(run_params[i][2])               #Sets nn_type parameter
    agent_goal_type=int(run_params[i][3])       #Sets agent_goal_type parameter
    env_learn_type=int(run_params[i][4])        #Sets env_learn_type parameter, use Sensed States or Sensed Deltas
    env_noise=float(run_params[i][5])           #Sets the level of environmental noise
    agent_move_type=int(run_params[i][6])       #Sets agent_move_type parameter


    #Create Run String
    run_str = ('py -3.7 RoboPers_Sim.py ' + str(local) + ' ' + str(congruent) + ' ' + str(nn_type) + ' ' + \
                   str(agent_goal_type) + ' ' + str(env_learn_type) + ' ' + str(env_noise) + ' ' + str(agent_move_type))
    
    #Set repeat cnt for probabilistic models (for averaging over sample)
    tot_reps=repeat=30


    #Run program, get output
    while repeat>0:
        print("*Repeats Left:", repeat)
        #Run python program    
        os.system(run_str)
        
        #Get output file
        temp_array=[]
        output_file= open('outfile.txt', 'r')
        output_file.readline()                                  #Throwaway header line
        out_line=output_file.readline().strip()
        temp_array = out_line.split('\t')                       #Main output line

        output_file.readline()                                  #Skip blank line
        out_line=output_file.readline().strip()
        prog_run_time = out_line.split('\t')[1]                 #Program Run Time 
        

        ##### Output Array Load ##########
        #Set Descriptors
        temp_array2=[]
        temp_array2.append(i)                                   #Just use the run order as "model sort" in this version

        if local==0:
            temp_array2.append('Global')
        elif local==1:
            temp_array2.append('Local')
        elif local==2:
            temp_array2.append('Super-Local')
            
        if congruent==0:
            temp_array2.append('Incongruent')
        elif congruent==1:
            temp_array2.append('Congruent')
        elif congruent==2:
            temp_array2.append('Random')

        if nn_type==1:
            temp_array2.append('MLP')
        elif nn_type==2:
            temp_array2.append('RNN')

        if agent_goal_type==0:
            temp_array2.append('None')
        elif agent_goal_type==1:
            temp_array2.append('Max Pos')
        elif agent_goal_type==2:
            temp_array2.append('Diff Neutral')

        if env_learn_type==1:
            temp_array2.append('States')
        elif env_learn_type==2:
            temp_array2.append('Deltas')

        temp_array2.append(env_noise)

        if agent_move_type==1:
            temp_array2.append('Direct Look')
        elif agent_move_type==2:
            temp_array2.append('Biased Saccades')


        #Insert Output Data
        for x in range(len(temp_array)):
            temp_array2.append(temp_array[x])

        #Insert Prog RunTime
        temp_array2.append(prog_run_time)

        #Insert Run Params
        run_param_str=('"'+ str(local) + ',' + str(congruent) + ',' + str(nn_type) + ',' + str(agent_goal_type) + ',' +  \
                       str(env_learn_type) + ',' + str(env_noise) + ',' + str(agent_move_type) + '"')
        temp_array2.append(run_param_str)

        #Insert Repeat Value
        temp_array2.append(repeat)
        
        #Finally, load it into output array
        output.append(temp_array2)
        
        ####Update Repeat Value############
        repeat-=1
    

        

##################################################################
#
# Write Results 
#
##################

#Open Results outfile
results_file=open("results.csv", 'w')

#Write Header
for i in range(field_cnt):
    results_file.write(str(header[i]) + ',')
results_file.write('\n')

#Write Data
for i in range(len(output)):
    for j in range(field_cnt):
        results_file.write(str(output[i][j]) + ',')
    results_file.write('\n')

results_file.close()


