#Robot Personality Sim Config File
#Contains module-wide shared global variables.
#Only Global Constants are declared here.
#For Switches/Arrays and switch-dependent constants, instantiation happens in main prod_sim module

import sys
import random       

#############################################################################
#
#User Input Program Parameters
# and module-wide Global Constants
#
#######################

'''NOTE: Some of these parameters might be overwritten by user input at run-time,
    see user input section of main code'''

#General Run-time Parameters
max_iter=1000#10000                                                   #Max number of iterations for simulation (main loop)
local=1                                                         #Flag for whether agent only considers local environment or not (0=global, 1=local, 2=super-local)
congruent=1                                                     #Determines if environment responds congruently or not (0=incongruent, 1=congruent, 2=random)
env_noise=0.05                                                  #Scales how much noise there is in environment change, between 0-1
env_step_size=0.1                                               #Size of environmental response
agent_move_type=1                                               #Flag to control how agent "moves": 1=direct look at focal point, 2=biased saccades
agent_goal_type=0                                               #Flag to control type of agent goal (0=No goal, 1=maximizing positive emot levels, 2=maximizing distance from neutral emot)
emot_rand=0                                                     #Switch to turn on completely random emotion
env_learn_type=1                                                #Flag to control environment info agent uses to learn (1=sensed current state, 2=sensed deltas over time) 

#Attention Parameters
dull_thresh=.01                                                 #Threshold for triggering dulling effect, should probably be something small (but non-zero)
att_span=3                                                      #Attention Span, i.e. max number of times in env deltas can be below threshold before dulling effect kicks in (equivalent to "time_out" in RobotFace)
att_decay_rate=2                                                #Attention Decay Rate, the lower this number the faster the decay (e.g. 2 means it will decay by half each timestep)

#Environment setup
env_height=6                                                    #This parameter and below set the environment size
env_width=6

#Neural Net parameters
nn_type=1                                                       #Determines type of neural net to use (1=MLP, 2=Recurrent NN)
dropout_rate=0.0                                                #Sets the dropout rate for neural network to prune edges (0.0 means no dropout)
layer_cnt=1                                                     #Sets number of hidden layers in NN


#Global Arrays
action_list=[0,1,2]




#############################################################################
#
#Testing outfile (for capturing line-by-line decision info
# for later analysis)
#
#######################

outfile_sp = open("outfile_specifics.txt", 'w')
outfile_nn_learning = open("outfile_nn_learning.txt", 'w')


#############################################################################
#
#Helper Functions
#
#######################

def find_min_idx(arr, actions):                            
    '''Needs a single-dimension array of values, returns index of smallest,
        breaks ties by picking rand'''
    min_idx=0
    curr_val=arr[0]
    for i in range(1,len(actions[1:])):                     
        if arr[i]<curr_val:
            min_idx=i
            curr_val=arr[i]
        elif arr[i]==curr_val:
            if min_idx==0:
                pass
            else:
                if random.randint(0,1)==1:
                    min_idx=i
                    curr_val=arr[i]

    return min_idx
